@@ -13,14 +13,13 @@ from .backends.comprenc import ComprencBackend
 from .backends.local import Backend as LocalBackend
 from .common import (inode_for_path, sha256_fh, get_path, get_backend_cachedir,
                      get_seq_no, is_mounted, get_backend, load_params,
-                     freeze_basic_mapping)
+                     save_params, time_ns)
 from .database import NoSuchRowError, Connection
 from .metadata import create_tables, dump_and_upload_metadata, download_metadata
 from .parse_args import ArgumentParser
 from os.path import basename
 import apsw
 import os
-import hashlib
 import re
 import shutil
 import itertools
@@ -156,12 +155,23 @@ class Fsck(object):
     def check_cache(self):
         """Commit uncommitted cache files"""
 
-        log.info("Checking cached objects...")
+        log.info("Checking for dirty cache objects...")
         if not os.path.exists(self.cachedir):
             return
+        candidates = os.listdir(self.cachedir)
 
-        for filename in os.listdir(self.cachedir):
-            self.found_errors = True
+        if sys.stdout.isatty():
+            stamp1 = 0
+        else:
+            stamp1 = float('inf')
+
+        for (i, filename) in enumerate(candidates):
+            stamp2 = time.time()
+            if stamp2 - stamp1 > 1:
+                sys.stdout.write('\r..processed %d/%d files (%d%%)..'
+                                 % (i, len(candidates), i/len(candidates)*100))
+                sys.stdout.flush()
+                stamp1 = stamp2
 
             match = re.match('^(\\d+)-(\\d+)$', filename)
             if match:
@@ -170,11 +180,30 @@ class Fsck(object):
             else:
                 raise RuntimeError('Strange file in cache directory: %s' % filename)
 
-            self.log_error("Committing block %d of inode %d to backend", blockno, inode)
-
+            # Calculate block checksum
             with open(os.path.join(self.cachedir, filename), "rb") as fh:
                 size = os.fstat(fh.fileno()).st_size
-                hash_ = sha256_fh(fh)
+                hash_should = sha256_fh(fh)
+            log.debug('%s has checksum %s', filename, hash_should)
+
+            # Check if stored block has same checksum
+            try:
+                block_id = self.conn.get_val('SELECT block_id FROM inode_blocks '
+                                             'WHERE inode=? AND blockno=?',
+                                             (inode, blockno,))
+                hash_is = self.conn.get_val('SELECT hash FROM blocks WHERE id=?',
+                                            (block_id,))
+            except NoSuchRowError:
+                hash_is = None
+            log.debug('Inode %d, block %d has checksum %s', inode, blockno,
+                      hash_is)
+            if hash_should == hash_is:
+                os.unlink(os.path.join(self.cachedir, filename))
+                continue
+
+            self.found_errors = True
+            self.log_error("Writing dirty block %d of inode %d to backend", blockno, inode)
+            hash_ = hash_should
 
             try:
                 (block_id, obj_id) = self.conn.get_row('SELECT id, obj_id FROM blocks WHERE hash=?', (hash_,))
@@ -218,7 +247,7 @@ class Fsck(object):
 
         log.info('Checking lost+found...')
 
-        timestamp = time.time()
+        now_ns = time_ns()
         try:
             (inode_l, name_id) = self.conn.get_row("SELECT inode, name_id FROM contents_v "
                                                    "WHERE name=? AND parent_inode=?", (b"lost+found", ROOT_INODE))
@@ -227,7 +256,7 @@ class Fsck(object):
             self.found_errors = True
             self.log_error("Recreating missing lost+found directory")
             inode_l = self.create_inode(mode=stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR,
-                                        atime=timestamp, ctime=timestamp, mtime=timestamp,
+                                        atime_ns=now_ns, ctime_ns=now_ns, mtime_ns=now_ns,
                                         refcount=1)
             self.conn.execute("INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)",
                               (self._add_name(b"lost+found"), inode_l, ROOT_INODE))
@@ -241,7 +270,7 @@ class Fsck(object):
             # We leave the old inode unassociated, so that it will be added
             # to lost+found later on.
             inode_l = self.create_inode(mode=stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR,
-                                        atime=timestamp, ctime=timestamp, mtime=timestamp,
+                                        atime_ns=now_ns, ctime_ns=now_ns, mtime_ns=now_ns,
                                         refcount=1)
             self.conn.execute('UPDATE contents SET inode=? WHERE name_id=? AND parent_inode=?',
                               (inode_l, name_id, ROOT_INODE))
@@ -501,8 +530,9 @@ class Fsck(object):
                     continue
                 self.moved_inodes.add(inode)
 
-                for (name, name_id, id_p) in self.conn.query('SELECT name, name_id, parent_inode '
-                                                             'FROM contents_v WHERE inode=?', (inode,)):
+                affected_entries = list(self.conn.query('SELECT name, name_id, parent_inode '
+                                                        'FROM contents_v WHERE inode=?', (inode,)))
+                for (name, name_id, id_p) in affected_entries:
                     path = get_path(id_p, self.conn, name)
                     self.log_error("File may lack data, moved to /lost+found: %s", to_str(path))
                     (lof_id, newname) = self.resolve_free(b"/lost+found", escape(path))
@@ -552,8 +582,9 @@ class Fsck(object):
                 continue
             self.moved_inodes.add(inode)
 
-            for (name, name_id, id_p) in self.conn.query('SELECT name, name_id, parent_inode '
-                                                         'FROM contents_v WHERE inode=?', (inode,)):
+            affected_entries = list(self.conn.query('SELECT name, name_id, parent_inode '
+                                                    'FROM contents_v WHERE inode=?', (inode,)))
+            for (name, name_id, id_p) in affected_entries:
                 path = get_path(id_p, self.conn, name)
                 self.log_error("File may lack data, moved to /lost+found: %s", to_str(path))
                 (lof_id, newname) = self.resolve_free(b"/lost+found", escape(path))
@@ -621,10 +652,10 @@ class Fsck(object):
                     (id_p, name) = self.resolve_free(b"/lost+found", ("block-%d" % id_).encode())
                     self.log_error("Block %d not referenced, adding as /lost+found/%s",
                                    id_, to_str(name))
-                    timestamp = time.time()
+                    now_ns = time_ns()
                     size = self.conn.get_val('SELECT size FROM blocks WHERE id=?', (id_,))
                     inode = self.create_inode(mode=stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR,
-                                              mtime=timestamp, atime=timestamp, ctime=timestamp,
+                                              mtime_ns=now_ns, atime_ns=now_ns, ctime_ns=now_ns,
                                               refcount=1, size=size)
                     self.conn.execute('INSERT INTO inode_blocks (inode, blockno, block_id) VALUES(?,?,?)',
                                       (inode, 0, id_))
@@ -646,36 +677,32 @@ class Fsck(object):
 
         log.info('Checking blocks (checksums)...')
 
-        for (block_id, obj_id) in self.conn.query('SELECT id, obj_id FROM blocks '
-                                                  'WHERE hash IS NULL'):
+        for (block_id, obj_id) in list(self.conn.query('SELECT id, obj_id FROM blocks '
+                                                       'WHERE hash IS NULL')):
             self.found_errors = True
-            self.log_error("No cached checksum for block %d, recomputing...", block_id)
+
+            # This should only happen when there was an error during upload,
+            # so the object must not have been stored correctly. We cannot
+            # just recalculate the hash for the block, because then we may
+            # be modifying the contents of the inode that refers to this
+            # block.
+            self.log_error("No checksum for block %d, removing from table...", block_id)
 
             # At the moment, we support only one block per object
             assert self.conn.get_val('SELECT refcount FROM objects WHERE id=?',
                                      (obj_id,)) == 1
 
-            def do_read(fh):
-                sha = hashlib.sha256()
-                while True:
-                    buf = fh.read(BUFSIZE)
-                    if not buf:
-                        break
-                    sha.update(buf)
-                return sha.digest()
-            hash_ = self.backend.perform_read(do_read, "s3ql_data_%d" % obj_id)
-
-            self.conn.execute('UPDATE blocks SET hash=? WHERE id=?',
-                              (hash_, block_id))
+            self.conn.execute('DELETE FROM blocks WHERE id=?', (block_id,))
+            self.conn.execute('DELETE FROM objects WHERE id=?', (obj_id,))
 
     def create_inode(self, mode, uid=os.getuid(), gid=os.getgid(),
-                     mtime=None, atime=None, ctime=None, refcount=None,
+                     mtime_ns=None, atime_ns=None, ctime_ns=None, refcount=None,
                      size=0):
         '''Create inode'''
 
-        id_ = self.conn.rowid('INSERT INTO inodes (mode,uid,gid,mtime,atime,ctime,'
+        id_ = self.conn.rowid('INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,'
                               'refcount,size) VALUES (?,?,?,?,?,?,?,?)',
-                              (mode, uid, gid, mtime, atime, ctime, refcount, size))
+                              (mode, uid, gid, mtime_ns, atime_ns, ctime_ns, refcount, size))
 
         return id_
 
@@ -946,8 +973,11 @@ class Fsck(object):
                         continue
                     self.moved_inodes.add(id_)
 
-                    for (name, name_id, id_p) in self.conn.query('SELECT name, name_id, parent_inode '
-                                                                 'FROM contents_v WHERE inode=?', (id_,)):
+                    # Copy the list, or we may pick up the same entry again and again
+                    # (first from the original location, then from lost+found)
+                    affected_entries = list(self.conn.query('SELECT name, name_id, parent_inode '
+                                                            'FROM contents_v WHERE inode=?', (id_,)))
+                    for (name, name_id, id_p) in affected_entries:
                         path = get_path(id_p, self.conn, name)
                         self.log_error("File may lack data, moved to /lost+found: %s", to_str(path))
                         (_, newname) = self.resolve_free(b"/lost+found", escape(path))
@@ -1050,7 +1080,7 @@ class ROFsck(Fsck):
         db = Connection(path + '.db')
         db.execute('PRAGMA journal_mode = WAL')
 
-        param = load_params(path + '.params')
+        param = load_params(path)
         super().__init__(None, None, param, db)
 
     def check(self):
@@ -1130,12 +1160,14 @@ def parse_args(args):
     parser.add_backend_options()
     parser.add_version()
     parser.add_storage_url()
-    parser.add_fatal_warnings()
 
     parser.add_argument("--batch", action="store_true", default=False,
                       help="If user input is required, exit without prompting.")
     parser.add_argument("--force", action="store_true", default=False,
                       help="Force checking even if file system is marked clean.")
+    parser.add_argument("--force-remote", action="store_true", default=False,
+                      help="Force use of remote metadata even when this would "
+                        "likely result in data loss.")
     options = parser.parse_args(args)
 
     return options
@@ -1164,7 +1196,7 @@ def main(args=None):
 
     if os.path.exists(cachepath + '.params'):
         assert os.path.exists(cachepath + '.db')
-        param = load_params(cachepath + '.params')
+        param = load_params(cachepath)
         if param['seq_no'] < seq_no:
             log.info('Ignoring locally cached metadata (outdated).')
             param = backend.lookup('s3ql_metadata')
@@ -1202,13 +1234,20 @@ def main(args=None):
               In the later case, waiting for a while should fix the problem, in the former case you
               should try to run fsck on the computer where the file system has been mounted most
               recently.
+
+              You may also continue and use whatever metadata is available in the backend.
+              However, in that case YOU MAY LOOSE ALL DATA THAT HAS BEEN UPLOADED OR MODIFIED
+              SINCE THE LAST SUCCESSFULL METADATA UPLOAD. Moreover, files and directories that
+              you have deleted since then MAY REAPPEAR WITH SOME OF THEIR CONTENT LOST.
               ''')))
 
-        print('Enter "continue" to use the outdated data anyway:',
+        print('Enter "continue, I know what I am doing" to use the outdated data anyway:',
               '> ', sep='\n', end='')
-        if options.batch:
+        if options.force_remote:
+            print('(--force-remote specified, continuing anyway)')
+        elif options.batch:
             raise QuietError('(in batch mode, exiting)', exitcode=41)
-        if sys.stdin.readline().strip() != 'continue':
+        elif sys.stdin.readline().strip() != 'continue, I know what I am doing':
             raise QuietError(exitcode=42)
 
         param['seq_no'] = seq_no
@@ -1256,8 +1295,7 @@ def main(args=None):
     param['seq_no'] += 1
     param['needs_fsck'] = True
     backend['s3ql_seq_no_%d' % param['seq_no']] = b'Empty'
-    with open(cachepath + '.params', 'wb') as fh:
-        fh.write(freeze_basic_mapping(param))
+    save_params(cachepath, param)
 
     fsck = Fsck(cachepath + '-cache', backend, param, db)
     fsck.check()
@@ -1283,8 +1321,7 @@ def main(args=None):
     param['last-modified'] = time.time()
 
     dump_and_upload_metadata(backend, db, param)
-    with open(cachepath + '.params', 'wb') as fh:
-        fh.write(freeze_basic_mapping(param))
+    save_params(cachepath, param)
 
     log.info('Cleaning up local metadata...')
     db.execute('ANALYZE')
@@ -1322,9 +1359,9 @@ def renumber_inodes(db):
                (CTRL_INODE,))
 
     log.info('..inodes..')
-    db.execute('INSERT INTO inodes (id,mode,uid,gid,mtime,atime,ctime,refcount,size,locked,rdev) '
+    db.execute('INSERT INTO inodes (id,mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size,locked,rdev) '
                'SELECT (SELECT rowid FROM inode_map WHERE inode_map.id = inodes_old.id), '
-               '       mode,uid,gid,mtime,atime,ctime,refcount,size,locked,rdev FROM inodes_old')
+               '       mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size,locked,rdev FROM inodes_old')
 
     log.info('..inode_blocks..')
     db.execute('INSERT INTO inode_blocks (inode, blockno, block_id) '
